import time
from dataclasses import dataclass
from typing import List, Tuple, Any, Callable, Union

from junit_xml import TestCase


class TestCaseCategories:
    FUNCTION = "function"
    FIXTURE = "fixture"


class JunitCaseException(BaseException):
    def __init__(self, exception: BaseException, *args: object) -> None:
        super().__init__(*args)
        self.exception: BaseException = exception


@dataclass
class CaseFailure:
    message: str
    output: str = ""
    type: str = ""

    def __getitem__(self, item):
        return self.__getattribute__(item)


@dataclass
class TestCaseData:
    case: TestCase
    _func: Callable
    _start_time: float
    parametrize: Union[None, List[Tuple[str, Any]]] = None
    _has_parent = False

    @property
    def name(self):
        return self._func.__name__

    @property
    def has_parent(self):
        return self._has_parent

    def set_fin_time(self):
        self.case.elapsed_sec = time.time() - self._start_time

    def get_case_key(self):
        if self.parametrize:
            return tuple(sorted(self.parametrize))
        return ()

    def set_parametrize(self, params: List[Tuple[str, Any]]):
        self.parametrize = params
        self.case.name += f'[{", ".join([f"{p[0]}={p[1]}" for p in params])}]'

    def set_parent(self, case_parent_name: str):
        self._has_parent = True
        self.case.classname = f"{self.case.classname}.{case_parent_name}"



from ._junit_fixture_test_case import JunitFixtureTestCase
from ._junit_test_case import CaseFailure, JunitTestCase, TestCaseCategories
from ._junit_test_suite import DuplicateSuiteError, JunitTestSuite, SuiteNotExistError

__all__ = [
    "JunitTestCase",
    "CaseFailure",
    "TestCaseCategories",
    "JunitFixtureTestCase",
    "JunitTestSuite",
    "SuiteNotExistError",
    "DuplicateSuiteError",
]



import re
from types import GeneratorType
from typing import Callable, List, Union

import decorator
from _pytest.mark import Mark
from _pytest.python import Function

from ._junit_test_case import JunitTestCase, TestCaseCategories
from ._junit_test_suite import JunitTestSuite


class JunitFixtureTestCase(JunitTestCase):
    """
    This class defines TestCase for pytest fixtures.
    When use, need to make sure that the fixture come first, for example:
        @pytest.fixture
        @JunitFixtureTestCase()
        def my_fixture()
            ...
    """

    _generator: Union[GeneratorType, None]

    AFTER_YIELD_EXCEPTION_MESSAGE_PREFIX = "[TEARDOWN EXCEPTION]"

    def __init__(self) -> None:
        super().__init__()
        self._generator = None

    def __call__(self, function: Callable) -> Callable:
        self._func = function

        def wrapper(_, *args, **kwargs):
            value = self._wrapper(function, *args, **kwargs)
            yield value
            try:
                if self._generator:
                    self._teardown_yield_fixture(self._generator)
            except BaseException as e:
                self._add_failure(e, self.AFTER_YIELD_EXCEPTION_MESSAGE_PREFIX)
                raise
            finally:
                JunitTestSuite.fixture_cleanup(self._case_data, self.get_suite_key())

        return decorator.decorator(wrapper, function)

    @classmethod
    def _teardown_yield_fixture(cls, it) -> None:
        """Execute the teardown of a fixture function by advancing the iterator
        after the yield and ensure the iteration ends (if not it means there is
        more than one yield in the function)."""
        try:
            next(it)
        except StopIteration:
            pass

    def _execute_wrapped_function(self, *args, **kwargs):
        generator = super()._execute_wrapped_function(*args, **kwargs)
        try:
            self._generator = generator
            return next(generator)
        except (StopIteration, TypeError):
            return None

    def _on_wrapper_end(self):
        """
        Fixtures executing before the test suite, due to that issue the suite can't collect the fixture case
        if there is an exception during it's execution. If exception occur, it call JunitTestSuite class method
        collect_all that trigger the suite to collect all cases and export them into xml
        :return: None
        """
        self._case_data.case.category = TestCaseCategories.FIXTURE

        super(JunitFixtureTestCase, self)._on_wrapper_end()
        if len(self._case_data.case.failures) > 0:
            JunitTestSuite.collect_all()

    def _get_suite(self) -> Union[Callable, None]:
        """
        Get suite function as unique key.
        This function handles also on case that the test suite function decorated with pytest.mark.parametrize
        :return: Wrapped Suite function instance
        """

        for f_locals in [
            stack_local
            for stack_local in self._stack_locals
            if "self" in stack_local and isinstance(stack_local["self"], Function)
        ]:
            func: Function = f_locals["self"]

            try:
                # if pytest.mark.parametrize exist, get actual function from class while ignoring the add parameters
                if hasattr(func, "own_markers") and len(func.own_markers) > 0:
                    mark_function = self._get_mark_function(func.own_markers, func)
                    if func.callspec and func.callspec.params:
                        if not self._case_data.parametrize:
                            self._case_data.set_parametrize(list(func.callspec.params.items()))

                    if mark_function:
                        return mark_function

                if func.cls and self._is_suite_exist(func.cls, func.name):
                    return getattr(func.cls, func.name).__wrapped__

                if func.cls is None and self._is_suite_exist(func.module, func.name):
                    return getattr(func.module, func.name).__wrapped__
            except AttributeError:
                pass

        return None

    @classmethod
    def _is_suite_exist(cls, obj: Union, func_name: str):
        func = getattr(obj, func_name)
        if hasattr(func, "__wrapped__"):
            return JunitTestSuite.is_suite_exist(func.__wrapped__)
        return False

    @staticmethod
    def _get_mark_function(own_markers: List[Mark], func: Function) -> Union[Callable, None]:
        """
        If mark parameterize decorate test suite with given fixture _get_mark_function is searching
        for all parametrize arguments and permutations.
        If mark type is parametrize, set self._parametrize with the current argument permutation.
        The representation of function name and arguments for N parametrize marks are as followed:
            func.name = some_func_name[param1-param2-...-paramN]
        :param own_markers: Pytest markers taken from func stack_locals
        :param func: wrapped function contained with pytest Function object
        :return: wrapped function itself
        """
        marks = [m for m in own_markers if m.name == "parametrize"]
        marks_count = len(marks)

        if marks_count == 0:
            return None

        params_regex = "-".join(["(.*?)"] * marks_count)
        args = list(re.compile(r"(.*?)\[{0}]".format(params_regex)).findall(func.name).pop())
        func_name = args.pop(0)

        params = list()
        for i in range(marks_count):
            params.append((marks[i].args[0], args[i]))

        if func.cls:
            return getattr(func.cls, func_name).__wrapped__
        else:
            return getattr(func.module, func_name).__wrapped__



import inspect
import traceback
from typing import Any, Callable, Dict, List, Union

from ._test_case_data import TestCaseData, TestCaseCategories, CaseFailure, JunitCaseException
from junit_xml import TestCase

from ._junit_decorator import JunitDecorator
from ._junit_test_suite import JunitTestSuite


class JunitTestCase(JunitDecorator):
    """
    JunitTestCase is a decorator that represents single TestCase.
    When the decorated function finished its execution, it's registered to the relevant JunitTestSuite.
    TestCase will fail (TestCase failure) only when exception occurs during execution.
    """

    _func: Union[Callable, None]
    _stack_locals: List[Dict[str, Any]]
    _case_data: Union[TestCaseData, None]

    def __init__(self) -> None:
        super().__init__()
        self._stack_locals = list()
        self._case_data = None

    @property
    def name(self):
        if self._func:
            return self._case_data.name

    def _on_wrapper_start(self, function):
        super()._on_wrapper_start(function)
        self._stack_locals = [frame_info.frame.f_locals for frame_info in inspect.stack()]
        case = TestCase(name=function.__name__, classname=self._get_class_name(), category=TestCaseCategories.FUNCTION)
        self._case_data = TestCaseData(_start_time=self._start_time, case=case, _func=function)

    def _add_failure(self, e: BaseException, message_prefix: str = ""):
        message = f"{message_prefix} {str(e)}" if message_prefix else str(e)
        failure = CaseFailure(message=message, output=traceback.format_exc(), type=e.__class__.__name__)
        self._case_data.case.failures.append(failure)

    def _on_exception(self, e: BaseException):
        self._add_failure(e)
        raise JunitCaseException(exception=e)

    def _on_wrapper_end(self):
        self._case_data.set_fin_time()
        JunitTestSuite.register_case(self._case_data, self.get_suite_key())

    def get_suite_key(self) -> Union[Callable, None]:
        """
        Get suite function as unique key.
        This function handles also on case that the test suite function decorated with pytest.mark.parametrize and
        collect its parameters and record them into the test case.
        :return: Wrapped Suite function instance
        """
        try:
            suite_func = self._get_suite()
            return suite_func
        except AttributeError:
            if JunitTestSuite.FAIL_ON_MISSING_SUITE:
                raise
        return None

    def _get_suite(self):
        suite_arguments = dict()
        suite_func = None

        for f_locals in [
            stack_local for stack_local in self._stack_locals if "function" in stack_local and "self" in stack_local
        ]:
            suite = f_locals["self"]
            if isinstance(suite, JunitTestCase) and f_locals["function"].__name__ != self.name:
                if not self._case_data.has_parent:
                    self._case_data.set_parent(f_locals["function"].__name__)

            if isinstance(suite, JunitTestSuite):
                suite_func = f_locals["function"]
                stack_keys = list(inspect.signature(suite_func).parameters.keys())
                for stack in self._stack_locals:
                    if all(key in stack for key in stack_keys):
                        suite_arguments = stack
                        break

                self.__set_parametrize(suite_func, suite_arguments)
                break

        return suite_func

    def __set_parametrize(self, suite_func: Callable, suite_arguments: Dict[str, Any]):
        """
        Grab and set current JunitTestCase parametrize values
        :param suite_func: Parametrize wrapped suite function
        :param suite_arguments: Arguments passed to the function
        :return:
        """
        if hasattr(suite_func, "pytestmark") and len(suite_arguments) > 0:
            params = list()
            marks = [m for m in suite_func.pytestmark if m.name == "parametrize"]

            for i in range(len(marks)):
                arg = marks[len(marks) - i - 1].args[0]
                params.append((arg, suite_arguments.get(arg)))
            self._case_data.set_parametrize(params)



import datetime
import inspect
import os
from pathlib import Path
from typing import Callable, ClassVar, Dict, List, Union

import pytest

from ._test_case_data import TestCaseCategories, TestCaseData, CaseFailure, JunitCaseException
from junit_xml import TestCase, TestSuite, to_xml_report_string

from ._junit_decorator import JunitDecorator

import traceback


class SuiteNotExistError(KeyError):
    """ Test Suite decorator name is not exists in suites poll """


class DuplicateSuiteError(KeyError):
    """ Test Suite decorator name is already exist in suites poll """


class JunitTestSuite(JunitDecorator):
    """
    JunitTestSuite is a decorator that waits for JunitTestCases decorators to register their testcases instances.
    After all test cases has finished their execution, the JunitTestSuite instance collect all registered TestCases
    and generate junit xml accordingly.
    JunitTestSuite represents single TestSuite.
    Generated report format: junit_<Class Name>_<Suite Function Name>_report.xml
    The usage of this function is regardless of pytest unittest packages.
    Default report path can be override if DEFAULT_REPORT_PATH_KEY environment variable is set
    """

    _junit_suites: ClassVar[Dict[Callable, "JunitTestSuite"]] = dict()
    _report_dir: Path
    _cases = List[TestCase]
    _func: Union[Callable, None]
    suite: Union[TestSuite, None]

    DEFAULT_REPORT_PATH_KEY = "JUNIT_REPORT_DIR"
    FAIL_ON_MISSING_SUITE_KEY = "FAIL_ON_MISSING_SUITE"
    SUITE_TEST_CASE_MESSAGE_PREFIX = "[SUITE EXCEPTION]"  # Exception on suite (without test case)

    XML_REPORT_FORMAT = "junit_{suite_name}_report{args}.xml"
    FAIL_ON_MISSING_SUITE = os.getenv(FAIL_ON_MISSING_SUITE_KEY, "False").lower() in ["true", "1", "yes", "y"]

    def __init__(self, report_dir: Path = None, custom_filename: str = None):
        """
        :param report_dir: Target directory, created if not exists
        :param custom_filename: If set, xml report will set to <exported_filename>.xml
        """
        super().__init__()
        self._report_dir = self.get_report_dir(report_dir)
        self._cases = list()
        self.suite = None
        self._timestamp = datetime.datetime.now()
        self._has_uncollected_fixtures = False
        self._custom_filename = custom_filename
        self._self_test_case = None

    def _on_call(self):
        self._register()

    def _on_wrapper_end(self):
        self.suite = TestSuite(
            name=f"{self._get_class_name()}_{self.name}", test_cases=self._get_cases(), timestamp=self._timestamp
        )
        self._export(self.suite)

    @classmethod
    def get_report_file_name(cls, suite_name: str, args: str = None, custom_filename: str = None):
        if custom_filename:
            return f"{custom_filename}.xml"
        if args:
            return cls.XML_REPORT_FORMAT.format(suite_name=suite_name, args=f"[{args}]")
        return cls.XML_REPORT_FORMAT.format(suite_name=suite_name, args="")

    @classmethod
    def get_suite(cls, suite_key: Callable) -> Union["JunitTestSuite", None]:
        """
        Return suite if exists, None otherwise
        :param suite_key: Wrapped function as cases key
        :return:
        """
        if cls.is_suite_exist(suite_key):
            return cls._junit_suites[suite_key]
        return None

    @classmethod
    def get_report_dir(cls, report_dir: Union[Path, None]) -> Path:
        if report_dir is None:
            return Path(os.getenv(cls.DEFAULT_REPORT_PATH_KEY, Path.cwd()))
        return report_dir

    @classmethod
    def collect_all(cls):
        """
        Collect all junit test suites reports from external source
        :return: None
        """
        for junit_suite in cls._junit_suites.values():
            cls._on_wrapper_end(self=junit_suite)

    @classmethod
    def is_suite_exist(cls, suite_func: Callable):
        return suite_func in cls._junit_suites

    @classmethod
    def register_case(cls, test_data, suite_func: Callable) -> None:
        """
        Register test case to the relevant test suite
        :param test_data: TestCaseData instance
        :param suite_func: Wrapped function as cases key
        :return: None
        """
        if cls.is_suite_exist(suite_func):
            cls._add_case(cls.get_suite(suite_func), test_data)
        else:
            if cls.FAIL_ON_MISSING_SUITE:
                raise SuiteNotExistError(f"Can't find suite named {suite_func} for {test_data} test case")

    def _register(self):
        if self._func in JunitTestSuite._junit_suites:
            raise DuplicateSuiteError(f"Suite {self.name} already exist")
        JunitTestSuite._junit_suites[self._func] = self

    def _get_cases(self):
        if self._self_test_case:
            return [data.case for data in self._cases] + [self._self_test_case]
        return [data.case for data in self._cases]

    def _add_case(self, test_data):
        if test_data.case.category == "fixture":
            self._has_uncollected_fixtures = True
        return self._cases.append(test_data)

    def _get_parametrize_values(self) -> str:
        """
        In case of pytest parametrize, this function collect and return parametrizes values
        :return: underscore separated parametrize values
        """
        values = ""
        if len(self._cases) == 0:
            return "_".join([str(tup[1]) for tup in self._get_parameterized_on_no_cases()])

        test_cases = [test_case_data for test_case_data in self._cases if test_case_data.parametrize is not None]
        if test_cases:
            marks = {c.get_case_key() for c in self._cases if len(c.get_case_key()) > 0}
            if marks:
                values = "_".join([str(tup[1]) for tup in marks.pop()])
        return values

    def _export(self, suite: TestSuite) -> None:
        """
        Export test suite to JUnit xml file
        :param suite: TestSuite to export
        :return: None
        """
        if len(suite.test_cases) == 0:
            return

        if not self._has_uncollected_fixtures:
            values = self._get_parametrize_values()
            path = self._report_dir.joinpath(self.get_report_file_name(
                suite_name=suite.name, args=values, custom_filename=self._custom_filename))
            xml_string = to_xml_report_string([suite])

            os.makedirs(self._report_dir, exist_ok=True)
            with open(path, "w") as f:
                f.write(xml_string)

            self.clear_cases()

    def clear_cases(self):
        """ Delete all cases from suite """
        self._cases = list()
        self.suite.test_cases = list()

    @classmethod
    def fixture_cleanup(cls, test_data, suite_func: Callable):
        """
        Collect junit suite after fixture yield
        :param test_data: TestCaseData
        :param suite_func: suite key
        :return: None
        """
        junit_suite = cls.get_suite(suite_func)
        if junit_suite and junit_suite._cases:
            first_fixture = junit_suite._cases[0].case
            if first_fixture == test_data.case:
                junit_suite._collect_yield()

    def _collect_yield(self):
        """
        Export suite if exist
        :return:
        """
        self._has_uncollected_fixtures = False
        self._export(self.suite)

    def _on_exception(self, e: BaseException):
        if isinstance(e, JunitCaseException):
            raise e.exception
        self._self_test_case = TestCase(name=self._func.__name__, classname=self._get_class_name(),
                                        category=TestCaseCategories.FUNCTION)
        data = TestCaseData(_start_time=self._start_time, case=self._self_test_case, _func=self._func)
        message = "[SUITE EXCEPTION] " + str(e)
        failure = CaseFailure(message=message, output=traceback.format_exc(), type=e.__class__.__name__)
        data.case.failures.append(failure)
        raise

    def _get_parameterized_on_no_cases(self):
        stack_locals = [frame_info.frame.f_locals for frame_info in inspect.stack()]
        for f_locals in [stack_local for stack_local in stack_locals if "pyfuncitem" in stack_local]:
            pytest_func = f_locals["pyfuncitem"]
            if hasattr(self._func, "pytestmark"):
                parameterized = [m.args[0] for m in self._func.pytestmark if m.name == "parametrize"]
                if isinstance(pytest_func, pytest.Function) and parameterized:
                    return list({k: v for k, v in pytest_func.funcargs.items() if k in parameterized}.items())

        return []



import inspect
import re
import time
from abc import ABC, abstractmethod
from typing import Any, Callable

import decorator


class JunitDecorator(ABC):
    def __init__(self) -> None:
        self._func = None
        self._start_time = None

    def __call__(self, function: Callable) -> Callable:
        """
        :param function: Decorated function
        :return: Wrapped function
        """
        self._func = function
        self._on_call()

        def wrapper(_, *args, **kwargs):
            return self._wrapper(function, *args, **kwargs)

        return decorator.decorator(wrapper, function)

    @property
    def name(self):
        return self._func.__name__

    def _wrapper(self, function: Callable, *args, **kwargs):
        value = None

        self._on_wrapper_start(function)
        try:
            value = self._execute_wrapped_function(*args, **kwargs)
        except BaseException as e:
            self._on_exception(e)
        finally:
            self._on_wrapper_end()
        return value

    def _get_class_name(self) -> str:
        """
        Get class name of which the decorated function contained in it.
        If class doesn't exists, it returns the module name
        :return: class or module name
        """
        module = inspect.getmodule(self._func)
        try:
            classname, _ = re.compile(r"(\w+)\.(\w+)\sat\s").findall(str(self._func))[0]
            return classname
        except IndexError:
            return module.__name__

    @abstractmethod
    def _on_wrapper_end(self) -> None:
        """
        Executed after execution finished (successfully or not)
        :return: None
        """

    def _on_call(self) -> None:
        """
        Executed on __call__ start.
        :return: None
        """

    def _execute_wrapped_function(self, *args, **kwargs) -> Any:
        """
        Execute wrapped function and return its value.
        Exceptions in this function will be caught by _on_exception
        :param args: Arguments passed to the function
        :param kwargs: Key arguments passed to the function
        :return: Wrapped function return value
        """
        return self._func(*args, **kwargs)

    def _on_exception(self, e: BaseException) -> None:
        """
        This function executed when exception is raised within the wrapped function
        :param e: Raised BaseException
        :return: None
        """
        raise

    def _on_wrapper_start(self, function) -> None:
        """
        This function executed when wrapper function starts
        :return: None
        """
        self._start_time = time.time()



import os
import sys

import setuptools

ROOT = os.path.dirname(__file__)

if sys.version_info < (3, 6, 0):
    sys.exit("Python 3.6.0 is the minimum required version")

with open(os.path.join(ROOT, "README.md")) as f:
    long_description = f.read()

with open(os.path.join(ROOT, "requirements.txt")) as f:
    requirements = f.readlines()

with open(os.path.join(ROOT, "test-requirements.txt")) as f:
    test_requirements = f.readlines()

setuptools.setup(
    name="junit-report",
    setup_requires=["vcversioner"],
    vcversioner={"vcs_args": ["git", "describe", "--tags", "--long"]},
    description="Export your test suites and cases to JUnit report using decorators",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/eliorerz/junit-report",
    author="Elior Erez",
    author_email="elior123@gmail.com",
    license="MIT",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    install_requires=requirements,
    tests_require=requirements + test_requirements,
    include_package_data=True,
    python_requires=">=3.6.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Testing",
    ],
)
